(*
 * Interpreter for F - Main Program
 *
 * 2000/02/02 Andreas Rossberg
 *)


structure Main =
  struct
    open TextIO

    fun errorToString(Elab.IdUnbound x)         = "Unbound identifier " ^ x
      | errorToString(Elab.IfConditionNotBool)  = "Condition must be boolean"
      | errorToString(Elab.IfBranchesMismatch)  = "Type mismatch in if branches"
      | errorToString(Elab.AppNoFunction)       = "Operator is not a function"
      | errorToString(Elab.AppArgumentMismatch) = "Invalid argument type"

    fun typeToString t				  = typeToString' false t
    and typeToString' b     (Syntax.Bool)	  = "bool"
      | typeToString' b     (Syntax.Int)	  = "int"
      | typeToString' false (Syntax.Arrow(t1,t2)) =
	    typeToString' true t1 ^ " -> " ^ typeToString' false t2
      | typeToString' true  (Syntax.Arrow(t1,t2)) =
	    "(" ^ typeToString' true t1 ^ " -> " ^ typeToString' false t2 ^ ")"

    fun valueToString(Value.Num 0,  Syntax.Bool)    = "false"
      | valueToString(Value.Num 1,  Syntax.Bool)    = "true"
      | valueToString(Value.Num n,  Syntax.Int)     = Int.toString n
      | valueToString(Value.Op _,   Syntax.Arrow _) = "fn"
      | valueToString(Value.Proc _, Syntax.Arrow _) = "fn"
      | valueToString _                             = "Runtime error"


    fun fromString s =
	let
	    val tree = Parser.parse(Lexer.lex s)
	    val t    = Elab.elab Type.base tree
	    val v    = Eval.eval Value.base tree
	in
	    print(valueToString(v,t) ^ " : " ^ typeToString t ^ "\n")
	end
	handle Lexer.Error  => print "Lexical error\n"
	     | Parser.Error => print "Syntax error\n"
	     | Elab.Error e => print(errorToString e ^ "\n")
	     | Eval.Error   => print "Runtime error\n"

    fun fromFile n =
	let
	    val f = openIn n
	    val s = inputAll f
	    val _ = closeIn f
	in
	    fromString s
	end


    fun interactive() =
	case (print "> "; inputLine TextIO.stdIn)
	  of ""   => ()
	   | "\n" => interactive()
	   | s    => (fromString s; interactive())


    fun main []    = interactive()
      | main files = List.app (fn f => (print(f ^ "> "); fromFile f)) files

    val _ = main(CommandLine.arguments())

  end
